import sys
from importlib.metadata import version
from importlib.resources import path as rpath
from datetime import datetime, timezone
from configparser import ConfigParser
from logging import Logger, getLogger

from .utils import get_expanded_path

log: Logger = getLogger(__name__)


DEFAULT_CONFIG_PATH = '$XDG_CONFIG_HOME/pyssg/config.ini'
VERSION = version('pyssg')


def __expand_all_paths(config: ConfigParser) -> None:
    log.debug('expanding all path options')
    for option in config.options('path'):
        path: str = config.get('path', option)
        config.set('path', option, get_expanded_path(path))


def __check_well_formed_config(config: ConfigParser) -> None:
    log.debug('checking that config file is well formed')
    default_config: ConfigParser = ConfigParser()
    with rpath('pyssg.plt', 'default.ini') as p:
        log.debug('reading config file "%s"', p)
        default_config.read(p)

    for section in default_config.sections():
        log.debug('checking section "%s"', section)
        if not config.has_section(section):
            log.error('config does not have section "%s"', section)
            sys.exit(1)
        for option in default_config.options(section):
            log.debug('checking option "%s"', option)
            if not config.has_option(section, option):
                log.error('config does not have option "%s" in section "%s"', option, section)
                sys.exit(1)


def get_parsed_config(path: str) -> ConfigParser:
    config: ConfigParser = ConfigParser()
    log.debug('reading config file "%s"', path)
    config.read(path)

    __check_well_formed_config(config)
    __expand_all_paths(config)

    # set other required options
    log.debug('setting extra config options')
    config.set('fmt', 'rss_date', '%%a, %%d %%b %%Y %%H:%%M:%%S GMT')
    config.set('fmt', 'sitemap_date', '%%Y-%%m-%%d')
    config.set('info', 'version', VERSION)
    config.set('info', 'rss_run_date', datetime.now(
        tz=timezone.utc).strftime(config.get('fmt', 'rss_date')))
    config.set('info', 'sitemap_run_date', datetime.now(
        tz=timezone.utc).strftime(config.get('fmt', 'sitemap_date')))

    return config



import os
import sys
from importlib.resources import path as rpath
from typing import Union
from configparser import ConfigParser
from logging import Logger, getLogger, DEBUG

from .utils import create_dir, copy_file, get_expanded_path
from .arg_parser import get_parsed_arguments
from .configuration import get_parsed_config, DEFAULT_CONFIG_PATH, VERSION
from .database import Database
from .builder import Builder

log: Logger = getLogger(__name__)


def main() -> None:
    args: dict[str, Union[str, bool]] = vars(get_parsed_arguments())

    if args['debug']:
        # need to modify the root logger specifically,
        #   as it is the one that holds the config
        #   (__name__ happens to resolve to pyssg in __init__)
        root_logger: Logger = getLogger('pyssg')
        root_logger.setLevel(DEBUG)
        for handler in root_logger.handlers:
            handler.setLevel(DEBUG)
        log.debug('changed logging level to DEBUG')

    if not len(sys.argv) > 1 or (len(sys.argv) == 2 and args['debug']):
        log.info('pyssg v%s - no arguments passed, --help for more', VERSION)
        sys.exit(0)

    if args['version']:
        log.info('pyssg v%s', VERSION)
        sys.exit(0)

    config_path: str = args['config'] if args['config'] else DEFAULT_CONFIG_PATH
    config_path = get_expanded_path(config_path)
    config_dir, _ = os.path.split(config_path)
    log.debug('checked config file path, final config path "%s"', config_path)

    if args['copy_default_config']:
        log.info('copying default config file')
        create_dir(config_dir)
        with rpath('pyssg.plt', 'default.ini') as p:
            copy_file(p, config_path)
        sys.exit(0)

    if not os.path.exists(config_path):
        log.error('config file does\'t exist in path "%s"; make sure'
                  ' the path is correct; use --copy-default-config if it\'s the'
                  ' first time if you haven\'t already', config_path)
        sys.exit(1)

    config: ConfigParser = get_parsed_config(config_path)
    config.set('info', 'debug', str(args['debug']))

    if args['init']:
        log.info('initializing the directory structure and copying over templates')
        create_dir(config.get('path', 'src'))
        create_dir(os.path.join(config.get('path', 'dst'), 'tag'), True)
        create_dir(config.get('path', 'plt'))
        files: list[str] = ('index.html',
                            'page.html',
                            'tag.html',
                            'rss.xml',
                            'sitemap.xml')
        log.debug('list of files to copy over: (%s)', ', '.join(files))
        for f in files:
            plt_file: str = os.path.join(config.get('path', 'plt'), f)
            with rpath('pyssg.plt', f) as p:
                copy_file(p, plt_file)
        sys.exit(0)

    if args['add_checksum_to_db']:
        log.info('adding checksum column to existing db')
        db_path: str = os.path.join(config.get('path', 'src'), '.files')
        db: Database = Database(db_path, config)
        # needs to be read_old instead of read
        db.read_old()
        db.write()

        sys.exit(0)

    if args['build']:
        log.info('building the html files')
        db_path: str = os.path.join(config.get('path', 'src'), '.files')
        db: Database = Database(db_path, config)
        db.read()

        builder: Builder = Builder(config, db)
        builder.build()

        db.write()
        sys.exit(0)



import os
import sys
import shutil
from hashlib import md5
from logging import Logger, getLogger

log: Logger = getLogger(__name__)


def get_file_list(path: str,
                  exts: list[str],
                  exclude: list[str]=None) -> list[str]:
    log.debug('retrieving file list in path "%s" that contain file'
              ' extensions (%s) except (%s)',
              path, ', '.join(exts),
              ', '.join(exclude if exclude is not None else []))
    out: list[str] = []
    for root, dirs, files in os.walk(path):
        if exclude is not None:
            log.debug('removing excludes from list')
            dirs[:] = [d for d in dirs if d not in exclude]

        for f in files:
            if f.endswith(tuple(exts)):
                stripped_f: str = os.path.join(root, f).replace(path, '')[1:]
                out.append(stripped_f)
                log.debug('added file "%s" without "%s" part: "%s"',
                          f, path, stripped_f)
            else:
                log.debug('ignoring file "%s" as it doesn\'t contain'
                          ' any of the extensions (%s)', f, ', '.join(exts))

    return out


def get_dir_structure(path: str,
                      exclude: list[str]=None) -> list[str]:
    log.debug('retrieving dir structure in path "%s" except (%s)',
              path, ', '.join(exclude if exclude is not None else []))
    out: list[str] = []
    for root, dirs, files in os.walk(path):
        if exclude is not None:
            log.debug('removing excludes from list')
            dirs[:] = [d for d in dirs if d not in exclude]

        for d in dirs:
            if root in out:
                out.remove(root)
                log.debug('removed dir "%s" as it already is in the list', root)
            joined_dir: str = os.path.join(root, d)
            out.append(joined_dir)
            log.debug('added dir "%s" to the list', joined_dir)

    log.debug('removing "%s" from all dirs in list', path)
    return [o.replace(path, '')[1:] for o in out]


def create_dir(path: str, p: bool=False, silent=False) -> None:
    try:
        if p:
            os.makedirs(path)
        else:
            os.mkdir(path)
        if not silent: log.info('created directory "%s"', path)
    except FileExistsError:
        if not silent: log.info('directory "%s" already exists, ignoring', path)


def copy_file(src: str, dst: str) -> None:
    if not os.path.exists(dst):
        shutil.copy2(src, dst)
        log.info('copied file "%s" to "%s"', src, dst)
    else:
        log.info('file "%s" already exists, ignoring', dst)


# as seen in SO: https://stackoverflow.com/a/1131238
def get_checksum(path: str) -> str:
    log.debug('calculating md5 checksum for "%s"', path)
    file_hash = md5()
    with open(path, "rb") as f:
        while chunk := f.read(4096):
            file_hash.update(chunk)

    return file_hash.hexdigest()


def get_expanded_path(path: str) -> None:
    log.debug('expanding path "%s"', path)
    expanded_path: str = os.path.normpath(os.path.expandvars(path))
    if '$' in expanded_path:
        log.error('"$" character found in expanded path "%s";'
                  ' could be due to non-existant env var.', expanded_path)
        sys.exit(1)
    log.debug('expanded path "%s" to "%s"', path, expanded_path)

    return expanded_path



import os
import sys
from datetime import datetime, timezone
from logging import Logger, getLogger
from configparser import ConfigParser

log: Logger = getLogger(__name__)


class Page:
    def __init__(self,
                 name: str,
                 ctime: float,
                 mtime: float,
                 html: str,
                 meta: dict,
                 config: ConfigParser):
        log.debug('initializing the page object with name "%s"', name)
        # initial data
        self.name: str = name
        self.ctimestamp: float = ctime
        self.mtimestamp: float = mtime
        self.content: str = html
        self.meta: dict = meta
        self.config: ConfigParser = config

        # data from self.meta
        self.title: str = ''
        self.author: str = ''
        self.cdatetime: datetime = None
        self.mdatetime: datetime = None
        self.summary: str = ''
        self.lang: str = 'en'
        self.tags: list[tuple[str]] = []

        # constructed
        self.url: str = ''
        self.image_url: str = ''
        self.cdate: str = ''
        self.cdate_list: str = ''
        self.cdate_list_sep: str = ''
        self.cdate_rss: str = ''
        self.cdate_sitemap: str = ''
        self.mdate: str = None
        self.mdate_list: str = None
        self.mdate_list_sep: str = None
        self.mdate_rss: str = ''
        self.mdate_sitemap: str = ''

        # later assigned references to next and previous pages
        self.next: Page = None
        self.previous: Page = None

        # also from self.meta, but for og metadata
        self.og: dict[str, str] = dict()


    def __lt__(self, other):
        return self.ctimestamp < other.ctimestamp


    def __get_mandatory_meta(self, meta: str) -> str:
        try:
            log.debug('parsing required metadata "%s"', meta)
            return self.meta[meta][0]
        except KeyError:
            log.error('failed to parse mandatory metadata "%s" from file "%s"',
                      meta, os.path.join(self.config.get('path', 'src'), self.name))
            sys.exit(1)


    # parses meta from self.meta, for og, it prioritizes,
    # the actual og meta
    def parse_metadata(self):
        log.debug('parsing metadata for file "%s"', self.name)
        self.title = self.__get_mandatory_meta('title')
        self.author = self.__get_mandatory_meta('author')
        self.summary = self.__get_mandatory_meta('summary')
        self.lang = self.__get_mandatory_meta('lang')

        log.debug('parsing timestamp')
        self.cdatetime = datetime.fromtimestamp(self.ctimestamp,
                                                 tz=timezone.utc)
        self.cdate = self.cdatetime.strftime(self.config.get('fmt', 'date'))
        self.cdate_list = self.cdatetime.strftime(self.config.get('fmt', 'list_date'))
        self.cdate_list_sep = self.cdatetime.strftime(self.config.get('fmt', 'list_sep_date'))
        self.cdate_rss = self.cdatetime.strftime(self.config.get('fmt', 'rss_date'))
        self.cdate_sitemap = \
        self.cdatetime.strftime(self.config.get('fmt', 'sitemap_date'))

        if self.mtimestamp != 0.0:
            log.debug('parsing modified timestamp')
            self.mdatetime = datetime.fromtimestamp(self.mtimestamp,
                                                     tz=timezone.utc)
            self.mdate = self.mdatetime.strftime(self.config.get('fmt', 'date'))
            self.mdate_list = self.mdatetime.strftime(self.config.get('fmt', 'list_date'))
            self.mdate_list_sep = self.mdatetime.strftime(self.config.get('fmt', 'list_sep_date'))
            self.mdate_rss = self.mdatetime.strftime(self.config.get('fmt', 'rss_date'))
            self.mdate_sitemap = \
            self.mdatetime.strftime(self.config.get('fmt', 'sitemap_date'))
        else:
            log.debug('not parsing modified timestamp, hasn\'t been modified')

        try:
            tags_only: list[str] = self.meta['tags']
            log.debug('parsing tags')
            tags_only.sort()

            for t in tags_only:
                self.tags.append((t,
                                  f'{self.config.get("url", "main")}/tag/@{t}.html'))
        except KeyError:
            log.debug('not parsing tags, doesn\'t have any')

        log.debug('parsing url')
        self.url = f'{self.config.get("url", "main")}/{self.name.replace(".md", ".html")}'
        log.debug('final url "%s"', self.url)

        log.debug('parsing image url')
        try:
            self.image_url = \
            f'{self.config.get("url", "static")}/{self.meta["image_url"][0]}'
        except KeyError:
            log.debug('using default image, no image_url metadata found')
            self.image_url = \
            f'{self.config.get("url", "static")}/{self.config.get("url", "default_image")}'
        log.debug('final image url "%s"', self.image_url)

        # if contains open graph elements
        try:
            # og_e = object graph entry
            og_elements: list[str] = self.meta['og']
            log.debug('parsing og metadata')
            for og_e in og_elements:
                kv: str = og_e.split(',', 1)
                if len(kv) != 2:
                    log.error('invalid og syntax for "%s", needs to be "k, v"', og_e)
                    sys.exit(1)

                k: str = kv[0].strip()
                v: str = kv[1].strip()

                log.debug('og element: ("%s", "%s")', k, v)
                self.og[k] = v
        except KeyError:
            log.debug('no og metadata found')



import os
import sys
from logging import Logger, getLogger
from configparser import ConfigParser

from .utils import get_checksum

log: Logger = getLogger(__name__)


# db class that works for both html and md files
class Database:
    __OLD_COLUMN_NUM: int = 4
    __COLUMN_NUM: int = 5

    def __init__(self, db_path: str,
                 config: ConfigParser):
        log.debug('initializing the page db on path "%s"', db_path)
        self.db_path: str = db_path
        self.config: ConfigParser = config
        self.e: dict[str, tuple[float, float, str, list[str]]] = dict()


    # updates the tags for a specific entry (file)
    #   file_name only contains the entry name (without the absolute path)
    def update_tags(self, file_name: str,
                    tags: list[str]) -> None:
        if file_name in self.e:
            log.debug('updating tags for entry "%s"', file_name)
            cts, mts, checksum, old_tags = self.e[file_name]
            log.debug('entry "%s" old content: (%s, %s, %s, (%s))',
                      file_name, cts, mts, checksum, ', '.join(old_tags))
            self.e[file_name] = (cts, mts, checksum, tags)
            log.debug('entry "%s" new content: (%s, %s, %s, (%s))',
                      file_name, cts, mts, checksum, ', '.join(tags))
        else:
            log.error('can\'t update tags for entry "%s",'
                      ' as it is not present in db', file_name)
            sys.exit(1)


    # returns a bool that indicates if the entry
    # was (includes new entries) or wasn't updated
    def update(self, file_name: str,
               remove: str=None) -> bool:
        log.debug('updating entry for file "%s"', file_name)
        # initial default values
        f: str = file_name
        tags: list[str] = []
        if remove is not None:
            f = file_name.replace(remove, '')
            log.debug('removed "%s" from "%s": "%s"', remove, file_name, f)

        # get current time, needs actual file name
        time: float = os.stat(file_name).st_mtime
        log.debug('modified time for "%s": %s', file_name, time)

        # calculate current checksum, also needs actual file name
        checksum: str = get_checksum(file_name)
        log.debug('current checksum for "%s": "%s"', file_name, checksum)

        # two cases, 1) entry didn't exist,
        # 2) entry has been mod and,
        # 3) entry hasn't been mod
        #1)
        if f not in self.e:
            log.debug('entry "%s" didn\'t exist, adding with defaults', f)
            self.e[f] = (time, 0.0, checksum, tags)
            return True

        old_time, old_mod_time, old_checksum, tags = self.e[f]
        log.debug('entry "%s" old content: (%s, %s, %s, (%s))',
                  f, old_time, old_mod_time, old_checksum, ', '.join(tags))

        # 2)
        if checksum != old_checksum:
            if old_mod_time == 0.0:
                log.debug('entry "%s" has been modified for the first'
                          ' time, updating', f)
            else:
                log.debug('entry "%s" has been modified, updating', f)
            self.e[f] = (old_time, time, checksum, tags)
            log.debug('entry "%s" new content: (%s, %s, %s, (%s))',
                      f, old_time, time, checksum, ', '.join(tags))
            return True
        # 3)
        else:
            log.debug('entry "%s" hasn\'t been modified', f)
            return False


    def write(self) -> None:
        log.debug('writing db')
        with open(self.db_path, 'w') as file:
            for k, v in self.e.items():
                log.debug('parsing row for page "%s"', k)
                t: str = None
                row: str = None
                if len(v[3]) == 0:
                    t = '-'
                else:
                    t = ','.join(v[3])

                row = f'{k} {v[0]} {v[1]} {v[2]} {t}'

                log.debug('writing row: "%s\\n"', row)
                file.write(f'{row}\n')


    def _db_path_exists(self) -> bool:
        log.debug('checking that "%s" exists or is a file', self.db_path)
        if not os.path.exists(self.db_path):
            log.warning('"%s" doesn\'t exist, will be'
                        ' created once process finishes,'
                        ' ignore if it\'s the first run', self.db_path)
            return False

        if not os.path.isfile(self.db_path):
            log.error('"%s" is not a file"', self.db_path)
            sys.exit(1)

        return True


    def _read_raw(self) -> list[str]:
        rows: list[str] = None
        with open(self.db_path, 'r') as file:
            rows = file.readlines()
        log.debug('db contains %d rows', len(rows))

        return rows


    def read_old(self) -> None:
        log.debug('reading db with old schema (%d columns)', self.__OLD_COLUMN_NUM)
        if not self._db_path_exists():
            log.error('db path "%s" desn\'t exist, --add-checksum-to-db should'
                      'only be used when updating the old db schema', self.db_path)
            sys.exit(1)

        rows: list[str] = self._read_raw()
        cols: list[str] = None
        # l=list of values in entry
        log.debug('parsing rows from db')
        for it, row in enumerate(rows):
            i: int = it + 1
            r: str = row.strip()
            log.debug('row %d content: "%s"', i, r)
            # (file_name, ctimestamp, mtimestamp, [tags])
            cols: tuple[str, float, float, list[str]] = tuple(r.split())
            col_num: int = len(cols)
            if col_num != self.__OLD_COLUMN_NUM:
                log.critical('row %d doesn\'t contain %s columns, contains %d'
                             ' columns: "%s"',
                             i, self.__OLD_COLUMN_NUM, col_num, r)
                sys.exit(1)

            t: list[str] = None
            if cols[3] == '-':
                t = []
            else:
                t = cols[3].split(',')
            log.debug('tag content: (%s)', ', '.join(t))
            file_path: str = os.path.join(self.config.get('path', 'src'), cols[0])
            checksum: str = get_checksum(file_path)
            log.debug('checksum for "%s": "%s"', file_path, checksum)

            self.e[cols[0]] = (float(cols[1]), float(cols[2]), checksum, t)



    def read(self) -> None:
        log.debug('reading db')
        if not self._db_path_exists():
            return

        rows: list[str] = self._read_raw()
        cols: list[str] = None
        # l=list of values in entry
        log.debug('parsing rows from db')
        for it, row in enumerate(rows):
            i: int = it + 1
            r: str = row.strip()
            log.debug('row %d content: "%s"', i, r)
            # (file_name, ctimestamp, mtimestamp, checksum, [tags])
            cols: tuple[str, float, float, str, list[str]] = tuple(r.split())
            col_num: int = len(cols)
            if col_num == self.__OLD_COLUMN_NUM:
                log.error('row %d contains %d columns: "%s"; this is probably'
                          ' because of missing checksum column, which is used'
                          ' now to also check if a file has changed. Rerun'
                          ' with flag --add-checksum-to-db to add the checksum'
                          ' column to the current db; if you did any changes'
                          ' since last timestamp in db, it won\'t update'
                          ' modification timestamp',
                          i, self.__OLD_COLUMN_NUM, r)
                sys.exit(1)

            if col_num != self.__COLUMN_NUM:
                log.critical('row %d doesn\'t contain %s columns, contains %d'
                             ' columns: "%s"',
                             i, self.__COLUMN_NUM, col_num, r)
                sys.exit(1)

            t: list[str] = None
            if cols[4] == '-':
                t = []
            else:
                t = cols[4].split(',')
            log.debug('tag content: (%s)', ', '.join(t))

            self.e[cols[0]] = (float(cols[1]), float(cols[2]), cols[3], t)



from argparse import ArgumentParser, Namespace


def get_parsed_arguments() -> Namespace:
    parser = ArgumentParser(prog='pyssg',
                            description='''Static Site Generator that parses
                            Markdown files into HTML files. For datetime
                            formats see:
                            https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes''')
    parser.add_argument('-v', '--version',
                        action='store_true',
                        help='''print program version''')
    parser.add_argument('-c', '--config',
                        # don't give a default here, as it would seem like
                        #   --config was passed
                        # default='$XDG_CONFIG_HOME/pyssg/config.ini',
                        type=str,
                        help='''config file (path) to read from; if not passed,
                        '$XDG_CONFIG_HOME/pyssg/config.ini' is used''')
    parser.add_argument('--copy-default-config',
                        action='store_true',
                        help='''copies the default config to path specified in
                        --config flag''')
    parser.add_argument('-i', '--init',
                        action='store_true',
                        help='''initializes the directory structures and copies
                        over default templates''')
    parser.add_argument('-b', '--build',
                        action='store_true',
                        help='''generates all HTML files by parsing MD files
                        present in source directory and copies over manually
                        written HTML files''')
    parser.add_argument('-f', '--force',
                        action='store_true',
                        help='''force building all pages and not only the
                        updated ones''')
    parser.add_argument('--debug',
                        action='store_true',
                        help='''change logging level from info to debug''')
    parser.add_argument('--add-checksum-to-db',
                        action='store_true',
                        help='''add checksum column to db entries''')
    # really not needed, too much bloat and case scenarios to check for,
    #   instead, just read from config file or default config file
    """
    parser.add_argument('-s', '--src',
                        default='src',
                        type=str,
                        help='''src directory; handmade files, templates and
                        metadata directory; defaults to 'src' ''')
    parser.add_argument('-d', '--dst',
                        default='dst',
                        type=str,
                        help='''dst directory; generated (and transfered html)
                        files; defaults to 'dst' ''')
    parser.add_argument('-t', '--plt',
                        default='plt',
                        type=str,
                        help='''plt directory; all template files; defaults to
                        'plt' ''')
    parser.add_argument('-u', '--url',
                        default='',
                        type=str,
                        help='''base url without trailing slash''')
    parser.add_argument('--static-url',
                        default='',
                        type=str,
                        help='''base static url without trailing slash''')
    parser.add_argument('--default-image-url',
                        default='',
                        type=str,
                        help='''default image url''')
    parser.add_argument('--title',
                        default='Blog',
                        type=str,
                        help='''general title for the website; defaults to
                        'Blog' ''')
    parser.add_argument('--date-format',
                        default='%a, %b %d, %Y @ %H:%M %Z',
                        type=str,
                        help='''date format used inside pages (for creation and
                        modification times, for example); defaults to '%%a, %%b
                        %%d, %%Y @ %%H:%%M %%Z' ('Tue, Mar 16, 2021 @ 02:46 UTC',
                        for example)''')
    parser.add_argument('--list-date-format',
                        default='%b %d',
                        type=str,
                        help='''date format used for page entries in a list;
                        defaults to '%%b %%d' ('Mar 16', for example)''')
    parser.add_argument('--list-sep-date-format',
                        default='%B %Y',
                        type=str,
                        help='''date format used for the separator between page
                        entries in a list; defaults to '%%B %%Y' ('March 2021',
                        for example)''')
    """

    return parser.parse_args()



from .pyssg import main


# since this is not used as a package, rather it's used as a command line tool,
#   this is never called because pyssg:main is called directly when running pyssg
if __name__ == '__main__':
    main()



from logging import Logger, StreamHandler, getLogger, INFO

from .pyssg import main
from .per_level_formatter import PerLevelFormatter


# since this is the root package, setup the logger here
__LOG_LEVEL: int = INFO
log: Logger = getLogger(__name__)
log.setLevel(__LOG_LEVEL)
ch: StreamHandler = StreamHandler()
ch.setLevel(__LOG_LEVEL)
ch.setFormatter(PerLevelFormatter())
log.addHandler(ch)

# not meant to be used as a package, so just give main
__all__ = ['main']


from logging import Formatter, DEBUG, INFO, WARNING, ERROR, CRITICAL


class PerLevelFormatter(Formatter):
    # colors for the terminal in ansi
    __YELLOW: str = "\x1b[33m"
    __RED: str = "\x1b[31m"
    __BOLD_RED: str = "\x1b[31;1m"
    __RESET: str = "\x1b[0m"

    __DATE_FMT: str = '%Y-%m-%d %H:%M:%S'
    __COMMON_FMT: str = '[%(levelname)s] [%(module)s:%(funcName)s:%(lineno)d]: %(message)s'
    __FORMATS: dict[int, str] = {
        DEBUG: __COMMON_FMT,
        INFO: '%(message)s',
        WARNING: f'{__YELLOW}{__COMMON_FMT}{__RESET}',
        ERROR: f'{__RED}{__COMMON_FMT}{__RESET}',
        CRITICAL: f'{__BOLD_RED}{__COMMON_FMT}{__RESET}'
    }


    def format(self, record: str) -> str:
        fmt: str = self.__FORMATS.get(record.levelno)
        formatter: Formatter = Formatter(
            fmt=fmt, datefmt=self.__DATE_FMT, style='%')

        return formatter.format(record)



import os
from copy import deepcopy
from operator import itemgetter
from configparser import ConfigParser
from logging import Logger, getLogger

from jinja2 import Environment, Template, FileSystemLoader as FSLoader

from .utils import get_file_list, get_dir_structure, create_dir, copy_file
from .database import Database
from .md_parser import MDParser
from .page import Page

log: Logger = getLogger(__name__)


class Builder:
    def __init__(self, config: ConfigParser,
                 db: Database):
        log.debug('initializing site builder')
        self.config: ConfigParser = config
        self.db: Database = db

        # the autoescape option could be a security risk if used in a dynamic
        # website, as far as i can tell
        log.debug('initializing the jinja environment')
        self.__loader: FSLoader = FSLoader(self.config.get('path', 'plt'))
        self.env: Environment = Environment(loader=self.__loader,
                                            autoescape=False,
                                            trim_blocks=True,
                                            lstrip_blocks=True)

        self.dirs: list[str] = None
        self.md_files: list[str] = None
        self.html_files: list[str] = None

        # files and pages are synoyms
        self.all_files: list[Page] = None
        self.updated_files: list[Page] = None
        self.all_tags: list[str] = None
        self.common_vars: dict = None


    def build(self) -> None:
        log.debug('building site')
        self.dirs = get_dir_structure(self.config.get('path', 'src'),
                                      ['templates'])
        self.md_files = get_file_list(self.config.get('path', 'src'),
                                      ['.md'],
                                      ['templates'])
        self.html_files = get_file_list(self.config.get('path', 'src'),
                                        ['.html'],
                                        ['templates'])

        self.__create_dir_structure()
        self.__copy_html_files()

        parser: MDParser = MDParser(self.md_files,
                                    self.config,
                                    self.db)
        parser.parse_files()

        # just so i don't have to pass these vars to all the functions
        self.all_files = parser.all_files
        self.updated_files = parser.updated_files
        self.all_tags = parser.all_tags

        # dict for the keyword args to pass to the template renderer
        log.debug('adding config, all_pages and all_tags to exposed vars for jinja')
        self.common_vars = dict(config=self.config,
                                all_pages=self.all_files,
                                all_tags=self.all_tags)

        self.__render_articles()
        self.__render_tags()
        self.__render_template('index.html', 'index.html', **self.common_vars)
        self.__render_template('rss.xml', 'rss.xml', **self.common_vars)
        self.__render_template('sitemap.xml', 'sitemap.xml', **self.common_vars)


    def __create_dir_structure(self) -> None:
        log.debug('creating dir structure')
        dir_path: str = None
        for d in self.dirs:
            dir_path = os.path.join(self.config.get('path', 'dst'), d)
            # using silent=True to not print the info create dir msgs for this
            create_dir(dir_path, True, True)


    def __copy_html_files(self) -> None:
        if len(self.html_files) > 0:
            log.debug('copying all html files')
        else:
            log.debug('no html files to copy')
        src_file: str = None
        dst_file: str = None

        for f in self.html_files:
            src_file = os.path.join(self.config.get('path', 'src'), f)
            dst_file = os.path.join(self.config.get('path', 'dst'), f)

            # only copy files if they have been modified (or are new)
            if self.db.update(src_file, remove=f'{self.config.get("path", "src")}/'):
                log.debug('file "%s" has been modified or is new, copying', f)
                copy_file(src_file, dst_file)
            else:
                if self.config.getboolean('other', 'force'):
                    log.debug('file "%s" hasn\'t been modified, but option force is set to true, copying anyways', f)
                    copy_file(src_file, dst_file)
                else:
                    log.debug('file "%s" hasn\'t been modified, ignoring', f)


    def __render_articles(self) -> None:
        log.debug('rendering html')
        article_vars: dict = deepcopy(self.common_vars)
        temp_files: list[Page] = None

        # check if only updated should be created
        if self.config.getboolean('other', 'force'):
            log.debug('all html will be rendered, force is set to true')
            temp_files = self.all_files
        else:
            log.debug('only updated or new html will be rendered')
            temp_files = self.updated_files

        for p in temp_files:
            log.debug('adding page to exposed vars for jinja')
            article_vars['page'] = p
            # actually render article
            self.__render_template("page.html",
                                   p.name.replace('.md','.html'),
                                   **article_vars)


    def __render_tags(self) -> None:
        log.debug('rendering tags')
        tag_vars: dict = deepcopy(self.common_vars)
        tag_pages: list[Page] = None
        for t in self.all_tags:
            log.debug('rendering tag "%s"', t[0])
            # clean tag_pages
            tag_pages = []
            log.debug('adding all pages that contain current tag')
            for p in self.all_files:
                if p.tags is not None and t[0] in list(map(itemgetter(0),
                                                           p.tags)):
                    log.debug('adding page "%s" as it contains tag "%s"',
                              p.name, t[0])
                    tag_pages.append(p)

            log.debug('adding tag and tag_pages to exposed vars for jinja')
            tag_vars['tag'] = t
            tag_vars['tag_pages'] = tag_pages

            # actually render tag page
            self.__render_template('tag.html',
                                   f'tag/@{t[0]}.html',
                                   **tag_vars)


    def __render_template(self, template_name: str,
                          file_name: str,
                          **template_vars) -> None:
        log.debug('rendering html "%s" with template "%s"',
                  file_name, template_name)
        template: Template = self.env.get_template(template_name)
        content: str = template.render(**template_vars)
        dst_path: str = os.path.join(self.config.get('path', 'dst'), file_name)

        log.debug('writing html file to path "%s"', dst_path)
        with open(dst_path, 'w') as f:
            f.write(content)



import os
from operator import itemgetter
from markdown import Markdown
from configparser import ConfigParser
from logging import Logger, getLogger

from markdown import Markdown
from yafg import YafgExtension
from MarkdownHighlight.highlight import HighlightExtension
from markdown_checklist.extension import ChecklistExtension

from .database import Database
from .page import Page

log: Logger = getLogger(__name__)


def _get_md_obj() -> Markdown:
    exts: list = ['extra',
                  'meta',
                  'sane_lists',
                  'smarty',
                  'toc',
                  'wikilinks',
                  # stripTitle generates an error when True,
                  # if there is no title attr
                  YafgExtension(stripTitle=False,
                                figureClass="",
                                figcaptionClass="",
                                figureNumbering=False,
                                figureNumberClass="number",
                                figureNumberText="Figure"),
                  HighlightExtension(),
                  ChecklistExtension(),
                  'pymdownx.caret',
                  'pymdownx.tilde']
    log.debug('list of md extensions: (%s)',
              ', '.join([e if isinstance(e, str) else type(e).__name__
                         for e in exts]))
    return Markdown(extensions=exts, output_format='html5')


# page and file is basically a synonym here...
class MDParser:
    def __init__(self, files: list[str],
                 config: ConfigParser,
                 db: Database):
        log.debug('initializing the md parser with %d files', len(files))
        self.files: list[str] = files

        self.config: ConfigParser = config
        self.db: Database = db
        self.md: Markdown = _get_md_obj()

        self.all_files: list[Page] = None
        # updated and modified are synonyms here
        self.updated_files: list[Page] = None
        self.all_tags: list[tuple[str]] = None


    def parse_files(self) -> None:
        log.debug('parsing all files')
        # initialize lists
        self.all_files = []
        self.updated_files = []
        self.all_tags = []
        # not used, not sure why i had this
        # all_tag_names: list[str] = []

        for f in self.files:
            log.debug('parsing file "%s"', f)
            src_file: str = os.path.join(self.config.get('path', 'src'), f)
            log.debug('path "%s"', src_file)
            # get flag if update is successful
            file_updated: bool = self.db.update(src_file, remove=f'{self.config.get("path", "src")}/')

            log.debug('parsing md into html')
            content: str = self.md.reset().convert(open(src_file).read())
            page: Page = Page(f,
                              self.db.e[f][0],
                              self.db.e[f][1],
                              content,
                              self.md.Meta,
                              self.config)
            page.parse_metadata()

            # keep a separated list for all and updated pages
            if file_updated:
                log.debug('has been modified, adding to mod file list')
                self.updated_files.append(page)
            log.debug('adding to file list')
            self.all_files.append(page)

            # parse tags
            if page.tags is not None:
                log.debug('parsing tags')
                # add its tag to corresponding db entry if existent
                self.db.update_tags(f, list(map(itemgetter(0), page.tags)))

                log.debug('add all tags to tag list')
                for t in page.tags:
                    if t[0] not in list(map(itemgetter(0), self.all_tags)):
                        log.debug('adding tag "%s" as it\'s not present in tag list', t[0])
                        self.all_tags.append(t)
                    else:
                        log.debug('ignoring tag "%s" as it\'s present in tag list', t[0])
            else:
                log.debug('no tags to parse')

        log.debug('sorting all lists for consistency')
        self.all_tags.sort(key=itemgetter(0))
        self.updated_files.sort(reverse=True)
        self.all_files.sort(reverse=True)

        pages_amount: int = len(self.all_files)
        # note that prev and next are switched because of the
        # reverse ordering of all_pages
        log.debug('update next and prev attributes')
        for i, p in enumerate(self.all_files):
            if i != 0:
                next_page: Page = self.all_files[i - 1]
                p.next = next_page

            if i != pages_amount - 1:
                prev_page: Page = self.all_files[i + 1]
                p.previous = prev_page
